//////////////////////////////////
//     //
//////////////////////////////////

#include "linkage.h"


//-----------------------------------------------------------------------------
//    
extern FILE* output;
//,            
extern bool sw_defh_ignore;


//-----------------------------------------------------------------------------
//    
char* str_new_copy(const char* source)
{
	char* dest = new char[strlen(source) + 1];
	return strcpy(dest, source);
}


//-----------------------------------------------------------------------------
//    
bool CCase::AddOneSpec(const CXMLEntry *xmlEntry, const char* ModuleName)
{
	//    
	const char* STag = xmlEntry->AttribValue("Tag");
	const char* SName = xmlEntry->AttribValue("Name");
	const char* SQualName = xmlEntry->AttribValue("Qual");
	if (!SQualName) SQualName = ModuleName;
	//  
	if (!SName) {
		//     
		if (!STag) {
			fprintf(output, "ERROR: Not found name in specialization in CASE \"%s\"\n", Name);
			return false;
		}
		//   (  NIL)
		SName = "";
	}
	//       
	if ((STag && tt_Type == TagType) || (!STag && tt_Type != TagType)) {
		fprintf(output, "ERROR: Wrong type of tag in CASE \"%s\" specialization \"%s\"", Name, SName);
		if (STag) fprintf(output, " with tag \"%s\"", STag);
		fprintf(output, "\n");
		return false;
	}
	//     (     )
	const char* AttribValueDefault = xmlEntry->AttribValue("Def");
	const bool is_default_spec = AttribValueDefault && !strcmp("yes", AttribValueDefault);
	if (is_default_spec && DefaultSpec) {
		fprintf(output, "ERROR: Default specialization redefined in CASE \"%s\"\n", Name);
		return false;
	}

	//     ( )
	bool spec_redef = false;
	//  
	TSpecStore::const_iterator ci;
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++) {
		bool is_incompatible;
		//   (    )
		if ((*ci)->IsEqual(STag, SQualName, SName, is_incompatible)) spec_redef = true;
		//  
		if (is_incompatible) goto incompatible_exit;
	}
	//   
	if (!spec_redef && DefaultSpec) {
		bool is_incompatible;
		//   (    )
		if(DefaultSpec->IsEqual(STag, SQualName, SName, is_incompatible)) spec_redef = true;
		//  
		if (is_incompatible) goto incompatible_exit;
	}
	//          
	if (spec_redef) return true;

	//         (   )
	if (is_default_spec) {
		DefaultSpec = new CSpec(STag, SQualName, SName);
		return true;
	}

	//    
	SpecStore.push_back(new CSpec(STag, SQualName, SName));
	return true;

incompatible_exit:
	//    
	fprintf(output, "ERROR: Incompatible specializations with tag \"%s\" in CASE \"%s\" declared in \"%s\"\n", STag, Name, QualName);
	return false;
}


//-----------------------------------------------------------------------------
//   
bool CCase::AddSpecs(const CXMLEntry* xmlEntry, const char* ModuleName)
{
	//    .2ml 
	CXMLEntry::TXMLEntryStore::const_iterator ci;
	for (ci = xmlEntry->XMLEntryStore.begin(); ci != xmlEntry->XMLEntryStore.end(); ci++)
		if (!strcmp("Spec", (*ci)->TagName))
			if (!AddOneSpec(*ci, ModuleName)) return false;
	return true;
}


//-----------------------------------------------------------------------------
//
CCase::CCase(const char* NewQualName, const char* NewName, bool is_decl, const CXMLEntry* xmlEntry) :
	DefaultSpec(NULL), is_declared(is_decl), QualName(NULL), Name(NULL), TagType(tt_Default)
{
	//  
	QualName = str_new_copy(NewQualName);
	Name = str_new_copy(NewName);
	//    
	const char* TT = xmlEntry->AttribValue("TagType");
	if (TT && !strcmp("type", TT)) TagType = tt_Type;
}


//-----------------------------------------------------------------------------
//
CCase::~CCase()
{
	//    
	delete[] QualName;
	delete[] Name;
	//  
	TSpecStore::const_iterator ci;
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++) delete *ci;
	//   
	delete DefaultSpec;
}


//-----------------------------------------------------------------------------
//   
const CSpec* CCase::FindSpecByTag(const char *Tag) const
{
	//  
	TSpecStore::const_iterator ci;
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++)
		if (!strcmp((*ci)->Tag, Tag)) return *ci;
	//    ( )
	if (DefaultSpec)
		if (!strcmp(DefaultSpec->Tag, Tag)) return DefaultSpec;
	//  
	return NULL;
}


//-----------------------------------------------------------------------------
//   
const CSpec* CCase::FindSpecByType(const char *sQual, const char *sName) const
{
	//,   sQual  (*ci)->QualName     
	//  
	TSpecStore::const_iterator ci;
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++)
		if (!strcmp(sQual, (*ci)->QualName) && !strcmp(sName, (*ci)->Name)) return *ci;
	//    ( )
	if (DefaultSpec)
		if (!strcmp(sQual, DefaultSpec->QualName) && !strcmp(sName, DefaultSpec->Name)) return DefaultSpec;
	return NULL;
}


//-----------------------------------------------------------------------------
//  
void CCase::WriteCPP(FILE *fh, FILE *fc) const
{
	//     
	fprintf(fh, "//CASE \"%s\" declared in module \"%s\"\n", Name, QualName);

	//             
	if (DefaultSpec) {
		SpecStore.push_back(DefaultSpec);
		DefaultSpec = NULL;
	}

	//      (  NIL)
	TSpecStore::const_iterator ci;
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++)
		if ('\0' != (*ci)->Name[0])
			fprintf(fh, "namespace %s {struct %s;}\n", (*ci)->QualName, (*ci)->Name);

	//  
	fprintf(fh, "namespace %s {\nstruct %s {\n", QualName, Name);
	//   : .  (0  
	//), .  -  
	fprintf(fh, "\tint O2M_SID;\n\tvoid (*O2M_FREE)(%s*);\n", Name);
	//   
	fprintf(fh, "\t%s() : O2M_SID(0), O2M_FREE(0) {}\n", Name);
	// ,    
	fprintf(fh, "\t%s(void (*O2M_PFUNC)(%s*)) {O2M_PFUNC(this);}\n", Name, Name);
	//   union     
	fprintf(fh, "\t~%s() {if (O2M_FREE) O2M_FREE(this);}\n\tunion {\n", Name);
	//  union (    NIL)
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++) {
		//  -  NIL    
		if ('\0' == (*ci)->Name[0])
			fprintf(fh, "\t\tvoid");
		else
			fprintf(fh, "\t\t%s::%s", (*ci)->QualName, (*ci)->Name);
		//    union
		fprintf(fh, "* O2M_SPEC_");
		WriteCPP_CompoundName(fh, *ci);
		fprintf(fh, ";\n");
	}
	fprintf(fh, "\t};\n");

	//         
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++) {
		//    -  .  
		fprintf(fh, "\t//<");
		WriteCPP_CompoundName(fh, *ci);
		fprintf(fh, ">\n\tstatic const int O2M_SID_");
		WriteCPP_CompoundName(fh, *ci);
		fprintf(fh, ";\n");
		//      
		fprintf(fh, "\tstatic void O2M_INIT_");
		WriteCPP_CompoundName(fh, *ci);
		fprintf(fh, "(%s*);\n", Name);
		//     ( -  NIL)
		if ('\0' != (*ci)->Name[0]) {
			fprintf(fh, "\tstatic void O2M_FREE_");
			WriteCPP_CompoundName(fh, *ci);
			fprintf(fh, "(%s*);\n", Name);
		}
	}
	fprintf(fh, "};}\n\n");

	//       
	int SID = 1;
	for (ci = SpecStore.begin(); ci != SpecStore.end(); ci++) {
		//    -  .  
		fprintf(fc, "\n//%s.%s<", QualName, Name);
		WriteCPP_CompoundName(fc, *ci);
		fprintf(fc, ">\nconst int %s::%s::O2M_SID_", QualName, Name);
		WriteCPP_CompoundName(fc, *ci);
		fprintf(fc, " = %i;\n", SID++);
		//      
		fprintf(fc, "\nvoid %s::%s::O2M_INIT_", QualName, Name);
		WriteCPP_CompoundName(fc, *ci);
		fprintf(fc, "(%s* O2M_PCASE)\n{\n", Name);
		if ('\0' != (*ci)->Name[0]) {
			//  NIL       
			fprintf(fc, "\tO2M_PCASE->O2M_SPEC_");
			WriteCPP_CompoundName(fc, *ci);
			fprintf(fc, " = new %s::%s;\n\tO2M_PCASE->O2M_FREE = O2M_FREE_", (*ci)->QualName, (*ci)->Name);
			WriteCPP_CompoundName(fc, *ci);
			fprintf(fc, ";\n");
		}
		fprintf(fc, "\tO2M_PCASE->O2M_SID = O2M_SID_");
		WriteCPP_CompoundName(fc, *ci);
		fprintf(fc, ";\n}\n");
		//     ( -  NIL)
		if ('\0' != (*ci)->Name[0]) {
			fprintf(fc, "\nvoid %s::%s::O2M_FREE_", QualName, Name);
			WriteCPP_CompoundName(fc, *ci);
			fprintf(fc, "(%s* O2M_PCASE)\n{\n\tdelete O2M_PCASE->O2M_SPEC_", Name);
			WriteCPP_CompoundName(fc, *ci);
			fprintf(fc, ";\n}\n");
		}
	}
	fprintf(fc, "\n");

}


//-----------------------------------------------------------------------------
//       
void CCase::WriteCPP_CompoundName(FILE *f, CSpec *Spec) const
{
	if (tt_Type == TagType)
		fprintf(f, "%s_%s", Spec->QualName, Spec->Name);
	else
		fprintf(f, "%s", Spec->Tag);
}


//-----------------------------------------------------------------------------
//
CCaseSumm::~CCaseSumm()
{
	//  
	TCaseStore::const_iterator ci;
	for (ci = CaseStore.begin(); ci != CaseStore.end(); ci++) delete *ci;
}


//-----------------------------------------------------------------------------
//   
bool CCaseSumm::CheckIntegrity() const
{
	TCaseStore::const_iterator ci;
	for (ci = CaseStore.begin(); ci != CaseStore.end(); ci++)
		if (!(*ci)->is_declared) {
			fprintf(output, "ERROR: Not found declaration of CASE \"%s\" in \"%s\"\n", (*ci)->Name, (*ci)->QualName);
			return false;
		};
	return true;
}


//-----------------------------------------------------------------------------
//    
CCase* CCaseSumm::FindCase(const char *QualName, const char *Name) const
{
	TCaseStore::const_iterator ci;
	for (ci = CaseStore.begin(); ci != CaseStore.end(); ci++)
		if (!strcmp(QualName, (*ci)->QualName) && !strcmp(Name, (*ci)->Name)) return *ci;
	return NULL;
}


//-----------------------------------------------------------------------------
//  
void CCaseSumm::WriteCPP(const CXMLEntry* xmlEntry, FILE *fh, FILE *fc) const
{
	//   ,    
	CXMLEntry::TXMLEntryStore::const_iterator cen;
	for (cen = xmlEntry->XMLEntryStore.begin(); cen != xmlEntry->XMLEntryStore.end(); ++cen) {
		//   XML -
		if (strcmp("File", (*cen)->TagName)) continue;
		//   XML - ( )
		const char* Path = (*cen)->AttribValue("Path");
		//       
		fprintf(fc, "#include \"%s%s.h\"\n", Path ? Path : "", (*cen)->AttribValue("Name"));
	}

	//  
	TCaseStore::const_iterator ci;
	for (ci = CaseStore.begin(); ci != CaseStore.end(); ci++)
		(*ci)->WriteCPP(fh, fc);
}


//-----------------------------------------------------------------------------
//     
bool CCaseSumm::AddCase(const CXMLEntry *xmlEntry, const char* ModuleName)
{
	//   
	const char* QualName = xmlEntry->AttribValue("Qual");
	const char* Name = xmlEntry->AttribValue("Name");
	//    ( QualName)
	bool is_decl = NULL == QualName;
	if (is_decl) QualName = ModuleName;
	//    
	if (!QualName || !Name) {
		if (!QualName) fprintf(output, "ERROR: Not found module name of CASE %s\n", is_decl ? "declaration" : "extension");
		if (!Name) fprintf(output, "ERROR: Not found name of CASE\n");
		return false;
	}
	//    
	TCaseStore::const_iterator ci;
	for (ci = CaseStore.begin(); ci != CaseStore.end(); ci++)
		if (!strcmp(QualName, (*ci)->QualName) && !strcmp(Name, (*ci)->Name)) {
			//    
			if (is_decl && (*ci)->is_declared) {
				fprintf(output, "ERROR: CASE \"%s\" in \"%s\" redeclared\n", Name, QualName);
				return false;
			}
			//       
			return (*ci)->AddSpecs(xmlEntry, ModuleName);
		}
	//    ,  
	CCase* Case = new CCase(QualName, Name, is_decl, xmlEntry);
	CaseStore.push_back(Case);
	return Case->AddSpecs(xmlEntry, ModuleName);
}


//-----------------------------------------------------------------------------
//
CSpec::CSpec(const char* NewTag, const char* NewQualName, const char* NewName) :
	Tag(NULL), QualName(NULL), Name(NULL)
{
	if (NewTag) Tag = str_new_copy(NewTag);
	QualName = str_new_copy(NewQualName);
	Name = str_new_copy(NewName);
}


//-----------------------------------------------------------------------------
//
CSpec::~CSpec()
{
	delete[] Tag;
	delete[] QualName;
	delete[] Name;
}


//-----------------------------------------------------------------------------
//       
//IsIncompatible -  
bool CSpec::IsEqual(const char *STag, const char *SQualName, const char *SName, bool &IsIncompatible)
{
	//  
	if (STag) {
		//   
		if (strcmp(Tag, STag)) {
			IsIncompatible = false;
			return false;
		} else {
			//    ( )
			IsIncompatible = strcmp(QualName, SQualName) || strcmp(Name, SName);
			if (IsIncompatible) return false;
		}
	} else
		IsIncompatible = false;
	//   
	if (!strcmp(QualName, SQualName) && !strcmp(Name, SName))
		return true;
	else
		return false;
}


//-----------------------------------------------------------------------------
//
CParam::CParam(const char* NewTag, const char* NewQualName, const char* NewName) :
	Tag(NULL), QualName(NULL), Name(NULL)
{
	if (NewTag) Tag = str_new_copy(NewTag);
	if (NewQualName) QualName = str_new_copy(NewQualName);
	if (NewName) Name = str_new_copy(NewName);
}


//-----------------------------------------------------------------------------
//
CParam::~CParam()
{
	delete[] Tag;
	delete[] QualName;
	delete[] Name;
}


//-----------------------------------------------------------------------------
//    (,     )
bool CParam::IsEqualTag(const CParam *Param) const
{
	if (Tag)
		return !strcmp(Tag, Param->Tag);
	else {
		if (!( !strcmp(Name, Param->Name) && ((QualName && Param->QualName) || (!QualName && !Param->QualName)) )) return false;
		if (QualName) return !strcmp(QualName, Param->QualName);
		return true;
	}
}


//-----------------------------------------------------------------------------
//      
bool CParam::IsEqualType(const CParam* Param) const
{
	return ((Tag && Param->Tag) || (!Tag && !Param->Tag)) && ((Tag && !Param->Name) || (!Tag && Param->Name));
}


//-----------------------------------------------------------------------------
//     
bool CCommonProc::AddOneCaseParam(const CXMLEntry* xmlEntry, const char* ModuleName)
{
	//     
	const char* prQualName = xmlEntry->AttribValue("Qual");
	const char* prName = xmlEntry->AttribValue("Name");
	//       
	if (!prQualName) prQualName = ModuleName;

	//     
	CaseParamStore.push_back(new CParam(NULL, prQualName, prName));
	return true;
}


//-----------------------------------------------------------------------------
//  
bool CCommonProc::AddCaseParams(const CXMLEntry* xmlEntry, const char* ModuleName)
{
	//     .2ml 
	CXMLEntry::TXMLEntryStore::const_iterator ci;
	for (ci = xmlEntry->XMLEntryStore.begin(); ci != xmlEntry->XMLEntryStore.end(); ci++)
		if (!strcmp("CPar", (*ci)->TagName))
			if (!AddOneCaseParam(*ci, ModuleName)) return false;
	//    1 
	if (CaseParamStore.empty()) {
		fprintf(output, "ERROR: Absent common parameters in common procedure \"%s\" in \"%s\"\n", Name, QualName);
		return false;
	}
	// 
	return true;
}


//-----------------------------------------------------------------------------
//
CCommonProc::CCommonProc(const char* NewQualName, const char* NewName, const bool NewDefH) :
	DefH(NewDefH), QualName(NULL), Name(NULL)
{
	//   
	QualName = str_new_copy(NewQualName);
	Name = str_new_copy(NewName);
}


//-----------------------------------------------------------------------------
//
CCommonProc::~CCommonProc()
{
	//    
	delete[] QualName;
	delete[] Name;
	//   
	TParamStore::const_iterator ci;
	for (ci = CaseParamStore.begin(); ci != CaseParamStore.end(); ci++) delete *ci;
}


//-----------------------------------------------------------------------------
//        
bool CCommonProc::CheckIntegrity(const THandlerProcStore* HPS, const CCaseSumm* CaseSumm) const
{
	//    
	const int sz = CaseParamStore.size();
	//    
	int HCount = 0;

	//     
	THandlerProcStore::const_iterator hci;
	for (hci = HPS->begin(); hci != HPS->end(); hci++)
		if (!strcmp(Name, (*hci)->Name) && !strcmp(QualName, (*hci)->QualName)) {
			//       
			if (sz != (*hci)->SpecParamStore.size()) {
				fprintf(output, "ERROR: Different counts of parameters in common procedure \"%s.%s\" and handler procedure \"%s\"[%s] in module \"%s\"\n", QualName, Name, (*hci)->Name, (*hci)->UID, (*hci)->ModuleName);
				return false;
			}
			//       
			for (int i = 0; i < sz; i++) {
				//      (    )
				CParam* Param = CaseParamStore[i];
				CCase* Case = CaseSumm->FindCase(Param->QualName, Param->Name);
				if (!Case) {
					fprintf(output, "ERROR: Unknown type of parameter N%i in common procedure \"%s.%s\"\n", i + 1, QualName, Name);
					return false;
				}
				// .    
				Param = (*hci)->SpecParamStore[i];
				//     
				if ((Param->Tag && CCase::tt_Type == Case->TagType) || (!Param->Tag && CCase::tt_Type != Case->TagType)) {
					fprintf(output, "ERROR: Wrong tag type in parameter N%i of handler procedure \"%s\"[%s] in module \"%s\"\n", i + 1, (*hci)->Name, (*hci)->UID, (*hci)->ModuleName);
					return false;
				}
				//    
				const CSpec* Spec;
				if (Param->Tag)
					Spec = Case->FindSpecByTag(Param->Tag);
				else
					Spec = Case->FindSpecByType(Param->QualName, Param->Name);
				//     
				if (!Spec) {
					fprintf(output, "ERROR: Unknown specialization in parameter N%i of handler procedure \"%s\"[%s] in module \"%s\"\n", i + 1, (*hci)->Name, (*hci)->UID, (*hci)->ModuleName);
					return false;
				}
			}
			//  
			++HCount;
		}

	//        
	int MaxHCount = 1;
	for (int i = 0; i < sz; i++) {
		CParam* Param = CaseParamStore[i];
		CCase* Case = CaseSumm->FindCase(Param->QualName, Param->Name);
		MaxHCount *= Case->GetSpecCount();
	}
	//     (   )
	if (HCount < MaxHCount) {
		bool is_error = !(DefH || sw_defh_ignore);
		fprintf(output, "%s: Insufficiency of handler procedures for common procedure \"%s.%s\"%s\n", is_error ? "ERROR" : "WARNING", QualName, Name, DefH ? "" : " without default handler");
		if (is_error) return false;
	}

	return true;
}


//-----------------------------------------------------------------------------
//   
void CCommonProc::WriteCPP(FILE* fh, FILE* fc, const THandlerProcStore* HPS, const CCaseSumm* CaseSumm) const
{
	// 
	const char* const textCommonComment = "//COMMON PROCEDURE \"%s\" declared in module \"%s\"\n";
	const char* const textCommonDecl = "void* O2M_COMMON_%s_%s(";

	//       
	fprintf(fh, textCommonComment, Name, QualName);
	fprintf(fc, textCommonComment, Name, QualName);

	//////////////////////////////////////////////////////////
	//      
	fprintf(fh, textCommonDecl, QualName, Name);
	fprintf(fc, textCommonDecl, QualName, Name);
	//  
	TParamStore::const_iterator ci;
	int num = 0;
	for (ci = CaseParamStore.begin();;) {
		fprintf(fh, "%s::%s* p%i", (*ci)->QualName, (*ci)->Name, num);
		fprintf(fc, "%s::%s* p%i", (*ci)->QualName, (*ci)->Name, num);
		++ci;
		if (CaseParamStore.end() == ci) break;
		fprintf(fh, ", ");
		fprintf(fc, ", ");
		++num;
	}
	//    
	fprintf(fh, ");\n\n");


	/////////////////////////////
	//   -
	fprintf(fc, ")\n{\n");

	//     
	THandlerProcStore::const_iterator hci;
	for (hci = HPS->begin(); hci != HPS->end(); hci++)
		if (!strcmp(Name, (*hci)->Name) && !strcmp(QualName, (*hci)->QualName)) {
			//     
			fprintf(fc, "\tif (");
			//  
			num = 0;
			for (ci = (*hci)->SpecParamStore.begin();;num++) {
				//      
				fprintf(fc, "p%i->O2M_SID_", num);
				//      O2M_SID
				if ((*ci)->Tag)
					fprintf(fc, "%s", (*ci)->Tag);
				else
					fprintf(fc, "%s_%s", (*ci)->QualName, (*ci)->Name);
				fprintf(fc, " == p%i->O2M_SID", num);
				//  . 
				++ci;
				if ((*hci)->SpecParamStore.end() == ci) break;
				fprintf(fc, " && ");
			}
			//   
			fprintf(fc, ") return (void*)%s::O2M_HANDLER_%s_%s;\n", (*hci)->ModuleName, (*hci)->Name, (*hci)->UID);
		}

	//   -
	fprintf(fc, "\treturn 0;\n}\n\n");
}


//-----------------------------------------------------------------------------
//     
bool CHandlerProc::AddOneSpecParam(const CXMLEntry* xmlEntry, const char* ModuleName)
{
	//     
	const char* prQualName = xmlEntry->AttribValue("Qual");
	const char* prName = xmlEntry->AttribValue("Name");
	const char* prTag = xmlEntry->AttribValue("Tag");
	//       
	if (!prQualName) prQualName = ModuleName;

	//     
	SpecParamStore.push_back(new CParam(prTag, prQualName, prName));
	return true;
}


//-----------------------------------------------------------------------------
//  
bool CHandlerProc::AddSpecParams(const CXMLEntry* xmlEntry, const char* ModuleName)
{
	//     .2ml 
	CXMLEntry::TXMLEntryStore::const_iterator ci;
	for (ci = xmlEntry->XMLEntryStore.begin(); ci != xmlEntry->XMLEntryStore.end(); ci++)
		if (!strcmp("SPar", (*ci)->TagName))
			if (!AddOneSpecParam(*ci, ModuleName)) return false;
	//    1 
	if (SpecParamStore.empty()) {
		fprintf(output, "ERROR: Absent common parameters in handler procedure \"%s\" in \"%s\"\n", Name, QualName);
		return false;
	}
	// 
	return true;
}


//-----------------------------------------------------------------------------
//
CHandlerProc::CHandlerProc(const char* NewQualName, const char* NewName, const char* NewUID, const char* NewModuleName) :
	ModuleName(NULL), Name(NULL), QualName(NULL), UID(NULL)
{
	//   UID 
	QualName = str_new_copy(NewQualName);
	Name = str_new_copy(NewName);
	UID = str_new_copy(NewUID);
	ModuleName = str_new_copy(NewModuleName);
}


//-----------------------------------------------------------------------------
//
CHandlerProc::~CHandlerProc()
{
	//    , UID
	delete[] ModuleName;
	delete[] Name;
	delete[] QualName;
	delete[] UID;
	//   
	TParamStore::const_iterator ci;
	for (ci = SpecParamStore.begin(); ci != SpecParamStore.end(); ci++) delete *ci;
}


//-----------------------------------------------------------------------------
//      
//(,   AddSpecParams  )
bool CHandlerProc::IsCompatible(const CHandlerProc *HP) const
{
	//   
	int sz = SpecParamStore.size();
	//   
	if (sz != HP->SpecParamStore.size()) return false;
	//  
	for (int i = 0; i < sz; i++)
		if (!SpecParamStore[i]->IsEqualType(HP->SpecParamStore[i])) return false;
	return true;
}


//-----------------------------------------------------------------------------
//      
//(,   AddSpecParams  IsCompatible  )
bool CHandlerProc::IsEqual(const CHandlerProc *HP) const
{
	//   
	int sz = SpecParamStore.size();
	//  
	for (int i = 0; i < sz; i++)
		if (!SpecParamStore[i]->IsEqualTag(HP->SpecParamStore[i])) return false;
	return true;
}


//-----------------------------------------------------------------------------
//   
bool CCommonProcSumm::AddCommonProc(const CXMLEntry* xmlEntry, const char* ModuleName)
{
	//    
	const char* QualName = ModuleName;
	const char* Name = xmlEntry->AttribValue("Name");
	const char* AttribValueDefH = xmlEntry->AttribValue("DefH");
	bool DefH = AttribValueDefH && !strcmp("yes", AttribValueDefH);
	//     
	if (!QualName || !Name) {
		if (!QualName) fprintf(output, "ERROR: Not found module name of common procedure\n");
		if (!Name) fprintf(output, "ERROR: Not found name of common procedure\n");
		return false;
	}
	//     
	TCommonProcStore::const_iterator ci;
	for (ci = CommonProcStore.begin(); ci != CommonProcStore.end(); ci++)
		if (!strcmp(QualName, (*ci)->QualName) && !strcmp(Name, (*ci)->Name)) {
			fprintf(output, "ERROR: Common procedure \"%s\" in \"%s\" redeclared\n", Name, QualName);
			return false;
		}
	//   
	CCommonProc* CommonProc = new CCommonProc(QualName, Name, DefH);
	CommonProcStore.push_back(CommonProc);
	return CommonProc->AddCaseParams(xmlEntry, ModuleName);
}


//-----------------------------------------------------------------------------
//    
bool CCommonProcSumm::AddHandlerProc(const CXMLEntry* xmlEntry, const char* ModuleName)
{
	//   
	const char* QualName = xmlEntry->AttribValue("Qual");
	if (!QualName) QualName = ModuleName;
	const char* Name = xmlEntry->AttribValue("Name");
	const char* UID = xmlEntry->AttribValue("UID");
	//    
	if (!QualName || !Name || !UID) {
		if (!QualName) fprintf(output, "ERROR: Not found module name of handler procedure\n");
		if (!Name) fprintf(output, "ERROR: Not found name of handler procedure\n");
		if (!UID) fprintf(output, "ERROR: Not found UID of handler procedure\n");
		return false;
	}
	//  
	CHandlerProc* HandlerProc = new CHandlerProc(QualName, Name, UID, ModuleName);
	bool res = HandlerProc->AddSpecParams(xmlEntry, ModuleName);
	if (!res) goto fault_exit;

	{//"{}" -   MinGW v3.1.0
		//         
		THandlerProcStore::const_iterator ci;
		for (ci = HandlerProcStore.begin(); ci != HandlerProcStore.end(); ci++)
			if (!strcmp(Name, (*ci)->Name) && !strcmp(QualName, (*ci)->QualName)) {
				//   
				if (!HandlerProc->IsCompatible(*ci)) {
					fprintf(output, "ERROR: Handler procedure \"%s.%s\"[%s] in \"%s\" and \"%s.%s\"[%s] in \"%s\" incompatible\n", QualName, Name, UID, ModuleName, (*ci)->QualName, (*ci)->Name, (*ci)->UID, (*ci)->ModuleName);
					goto fault_exit;
				}
				//       
				if (HandlerProc->IsEqual(*ci)) {
					fprintf(output, "ERROR: Handler procedure \"%s.%s\"[%s] in \"%s\" overload \"%s.%s\"[%s] in \"%s\"\n", QualName, Name, UID, ModuleName, (*ci)->QualName, (*ci)->Name, (*ci)->UID, (*ci)->ModuleName);
					goto fault_exit;
				}
			}
	}//"{}" -   MinGW v3.1.0

	//    
	HandlerProcStore.push_back(HandlerProc);
	return true;

fault_exit:
	//   
	delete HandlerProc;
	return false;
}


//-----------------------------------------------------------------------------
//
CCommonProcSumm::~CCommonProcSumm()
{
	//   
	TCommonProcStore::const_iterator cci;
	for (cci = CommonProcStore.begin(); cci != CommonProcStore.end(); cci++) delete *cci;
	//  
	THandlerProcStore::const_iterator hci;
	for (hci = HandlerProcStore.begin(); hci != HandlerProcStore.end(); hci++) delete *hci;
}


//-----------------------------------------------------------------------------
//      
bool CCommonProcSumm::CheckIntegrity(const CCaseSumm* CaseSumm) const
{
	//   
	TCommonProcStore::const_iterator cci;
	for (cci = CommonProcStore.begin(); cci != CommonProcStore.end(); cci++)
		if (!(*cci)->CheckIntegrity(&HandlerProcStore, CaseSumm)) return false;
	return true;
}


//-----------------------------------------------------------------------------
//          
void CCommonProcSumm::WriteCPP(FILE* fh, FILE* fc, const CCaseSumm* CaseSumm)
{
	//   
	TCommonProcStore::const_iterator ci;
	for (ci = CommonProcStore.begin(); ci != CommonProcStore.end(); ci++)
		(*ci)->WriteCPP(fh, fc, &HandlerProcStore, CaseSumm);
}

