//////////////////////////////////
//     //
//////////////////////////////////

#ifndef linkage_h_file
#define linkage_h_file

#include <vector>

#include "xml_lib.h"


//-----------------------------------------------------------------------------
//       
class CSpec {
public:
	CSpec(const char* NewTag, const char* NewQualName, const char* NewName);
	~CSpec();
	bool IsEqual(const char* STag, const char* SQualName, const char* SName, bool &IsIncompatible);
	// (    CASE TYPE)
	char* Tag;
	//    (  )
	char* QualName;
	char* Name;
};


//-----------------------------------------------------------------------------
//       
class CCase {
private:
	//    
	bool AddOneSpec(const CXMLEntry* xmlEntry, const char* ModuleName);
	//       
	void WriteCPP_CompoundName(FILE *f, CSpec* Spec) const;
	//  
	mutable CSpec* DefaultSpec;
	//  
	typedef std::vector<CSpec*> TSpecStore;
	// 
	mutable TSpecStore SpecStore;
public:
	//   
	bool AddSpecs(const CXMLEntry* xmlEntry, const char* ModuleName);
	//
	CCase(const char* NewQualName, const char* NewName, bool is_decl, const CXMLEntry* xmlEntry);
	~CCase();
	const CSpec* FindSpecByTag(const char* Tag) const;
	const CSpec* FindSpecByType(const char* sQual, const char* sName) const;
	//  - 
	int GetSpecCount() const {return SpecStore.size() + (DefaultSpec ? 1 : 0);};
	//  
	void WriteCPP(FILE* fh, FILE* fc) const;
	//    (    )
	bool is_declared;
	// ,  
	char* QualName;
	// 
	char* Name;
	//  
	enum {tt_Default, tt_Type} TagType;
};


//-----------------------------------------------------------------------------
//      
class CCaseSumm {
public:
	//     
	bool AddCase(const CXMLEntry* xmlEntry, const char* ModuleName);
	~CCaseSumm();
	bool CheckIntegrity() const;
	CCase* FindCase(const char* QualName, const char* Name) const;
	void WriteCPP(const CXMLEntry* xmlEntry, FILE* fh, FILE* fc) const;
private:
	//  
	typedef std::vector<CCase*> TCaseStore;
	// 
	TCaseStore CaseStore;
};


//-----------------------------------------------------------------------------
//       
class CParam {
public:
	CParam(const char* NewTag, const char* NewQualName, const char* NewName);
	~CParam();
	bool IsEqualTag(const CParam* Param) const;
	bool IsEqualType(const CParam* Param) const;
	// (    CASE TYPE)
	char* Tag;
	//   
	char* QualName;
	//     (/) 
	//()   ,  
	char* Name;
};


//-----------------------------------------------------------------------------
//  
typedef std::vector<CParam*> TParamStore;
//    
class CHandlerProc;
typedef std::vector<CHandlerProc*> THandlerProcStore;


//-----------------------------------------------------------------------------
//        
class CCommonProc {
private:
	//     
	bool AddOneCaseParam(const CXMLEntry* xmlEntry, const char* ModuleName);
	//  
	TParamStore CaseParamStore;
	//    
	bool DefH;
public:
	//  
	bool AddCaseParams(const CXMLEntry* xmlEntry, const char* ModuleName);
	//
	CCommonProc(const char* NewQualName, const char* NewName, const bool NewDefH);
	~CCommonProc();
	bool CheckIntegrity(const THandlerProcStore* HPS, const CCaseSumm* CaseSumm) const;
	void WriteCPP(FILE* fh, FILE* fc, const THandlerProcStore* HPS, const CCaseSumm* CaseSumm) const;
	// ,   
	char* QualName;
	//  
	char* Name;
};


//-----------------------------------------------------------------------------
//         
class CHandlerProc {
private:
	//     
	bool AddOneSpecParam(const CXMLEntry* xmlEntry, const char* ModuleName);
public:
	//  
	bool AddSpecParams(const CXMLEntry* xmlEntry, const char* ModuleName);
	//
	CHandlerProc(const char* NewQualName, const char* NewName, const char* NewUID, const char* NewModuleName);
	~CHandlerProc();
	bool IsCompatible(const CHandlerProc* HP) const;
	bool IsEqual(const CHandlerProc* HP) const;
	//  
	TParamStore SpecParamStore;
	// ,    
	char* ModuleName;
	// 
	char* Name;
	// ,   ,    
	char* QualName;
	//     
	char* UID;
};


//-----------------------------------------------------------------------------
//       
class CCommonProcSumm {
public:
	//   
	bool AddCommonProc(const CXMLEntry* xmlEntry, const char* ModuleName);
	//    
	bool AddHandlerProc(const CXMLEntry* xmlEntry, const char* ModuleName);
	~CCommonProcSumm();
	bool CheckIntegrity(const CCaseSumm* CaseSumm) const;
	void WriteCPP(FILE* fh, FILE* fc, const CCaseSumm* CaseSumm);
private:
	//   
	typedef std::vector<CCommonProc*> TCommonProcStore;
	//  
	TCommonProcStore CommonProcStore;
	//   
	THandlerProcStore HandlerProcStore;
};


#endif
